from __future__ import absolute_import, unicode_literals

REFBASE = 'http://docs.celeryproject.org/en/latest'
REFS = {
    'mailing-list':
        'https://groups.google.com/group/celery-users',
    'irc-channel': 'getting-started/resources.html#irc',
    'breakpoint-signal': 'tutorials/debugging.html',
    'internals-guide': 'internals/guide.html',
    'bundles': 'getting-started/introduction.html#bundles',
    'reporting-bugs': 'contributing.html#reporting-bugs',
}
